#include <stdio.h>
#include <stdlib.h>
#include <time.h>

static FILE *fp_trace;
static const char * output_trace_filename = "trace.out";

/*
 * The following attributes and prototypes are important:
 *
 * 'constructor' / 'destructor' is required by GCC to initiate tracing.
 * 
 * 'no_instrument_function' is optional, but important, otherwise GCC
 * will trace our tracing code, which easily leads to infinite recursion
 * and will cause the program to segfault.
 */
void __attribute__ ((constructor,no_instrument_function)) trace_begin (void);
void __attribute__ ((destructor,no_instrument_function)) trace_end (void);
void __attribute__ ((no_instrument_function)) __cyg_profile_func_enter (void *func, void *caller);
void __attribute__ ((no_instrument_function)) __cyg_profile_func_exit (void *func, void *caller);

void __attribute__ ((constructor,no_instrument_function)) trace_begin (void)
{ 
    printf("program start\n");

    fp_trace = fopen(output_trace_filename, "w"); 

    if(fp_trace == NULL) {
	    // perror("fopen");
	    printf("Failed to open output trace file: %s\n", output_trace_filename);
     }
}

void __attribute__ ((destructor,no_instrument_function)) trace_end (void) 
{ 
    if(fp_trace != NULL) { fclose(fp_trace); } 
    printf("program end\n");
}

void __attribute__ ((no_instrument_function)) __cyg_profile_func_enter (void *func, void *caller)
{ 
    fprintf(fp_trace, "e %p %p\n", func, caller);
}

void __attribute__ ((no_instrument_function)) __cyg_profile_func_exit (void *func, void *caller)
{ 
    fprintf(fp_trace, "x %p %p\n", func, caller);
}
